# SPDX-FileCopyrightText: 2025 René de Hesselle <dehesselle@web.de>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import logging
from pathlib import Path
from shutil import copytree

from pydantic_xml import BaseXmlModel

log = logging.getLogger("framework")


class Framework(BaseXmlModel):
    source_path: Path

    def install(self, bundle_dir: Path, install_prefix: Path):
        target_dir = bundle_dir / "Contents" / "Frameworks"

        if not target_dir.exists():
            log.debug(f"creating {target_dir}")
            target_dir.mkdir(parents=True)

        if (source_path := install_prefix / self.source_path).exists():
            target_path = target_dir / self.source_path.name
            if target_path.exists():
                log.error(f"will not overwrite {target_path}")
            else:
                log.debug(f"copy {source_path} to {target_path}")
                copytree(source_path, target_path, symlinks=True)
        else:
            log.error(f"cannot locate {self.source_path}")



# SPDX-FileCopyrightText: 2025 René de Hesselle <dehesselle@web.de>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import logging
from pathlib import Path
from typing import List

from pydantic_xml import BaseXmlModel, element

from .library import Library

log = logging.getLogger("library")


class Libraries(BaseXmlModel):
    libraries: List[Library] = element(tag="library")

    def install(self, bundle_dir: Path, install_prefix: Path):
        for library in self.libraries:
            library.install(bundle_dir, install_prefix)



# SPDX-FileCopyrightText: 2025 René de Hesselle <dehesselle@web.de>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from pathlib import Path
import logging
import os


class ExitOnCriticalHandler(logging.StreamHandler):
    def emit(self, record):
        if record.levelno == logging.CRITICAL:
            raise SystemExit(1)


class CollectStatisticsHandler(logging.StreamHandler):
    message_counter: dict[int, int] = dict()

    def __init__(self):
        super().__init__()
        self.messages = dict()
        for level in (
            logging.DEBUG,
            logging.INFO,
            logging.WARNING,
            logging.ERROR,
            logging.CRITICAL,
        ):
            self.messages[level] = 0

    def emit(self, record):
        try:
            self.messages[record.levelno] += 1
        except KeyError:
            self.messages[record.levelno] = 1

    @property
    def errors(self) -> int:
        return self.messages[logging.ERROR]

    @property
    def warnings(self) -> int:
        return self.messages[logging.WARNING]


class CustomLogRecord(logging.LogRecord):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.function_at_module = f"{self.name}:{self.funcName}"


logstats = CollectStatisticsHandler()


def setup_logging(logfile: Path) -> None:
    level = os.environ.get("ABCREATE_LOGLEVEL", "INFO").upper()

    file_handler = logging.FileHandler(logfile)
    file_handler.setLevel(level)
    file_handler.setFormatter(
        logging.Formatter(
            "%(asctime)-23s | %(name)-14s | %(funcName)-20s | %(levelname)-8s | %(message)s"
        )
    )
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(level)
    stream_handler.setFormatter(
        logging.Formatter(
            "%(asctime)-23s [%(function_at_module)-18s] %(levelname)s: %(message)s"
        )
    )
    logging.setLogRecordFactory(CustomLogRecord)
    logging.basicConfig(
        level=level,
        handlers=[
            file_handler,
            stream_handler,
            logstats,
            ExitOnCriticalHandler(),
        ],
    )



# SPDX-FileCopyrightText: 2025 René de Hesselle <dehesselle@web.de>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import logging
from pathlib import Path

from pydantic_xml import BaseXmlModel

from abcreate.bundle.library import Library
from abcreate.bundle.resource import Resource

log = logging.getLogger("glib")


class Glib(BaseXmlModel):
    def install(self, bundle_dir: Path, source_dir: Path):
        library = Library(source_path=Path("libglib-2.0.0.dylib"))
        library.install(bundle_dir, source_dir)

        for source_path in Path(source_dir / "lib" / "gio" / "modules").glob("*.so"):
            library = Library(source_path=source_path)
            # We cannot flatten here because GIO_MODULE_DIR expects a dedicated
            # directory (it will try to load everything in there, throwing warning
            # for libraries which are not GIO modules).
            library.install(bundle_dir, source_dir)

        resource = Resource(source_path=Path("share/glib-2.0"))
        resource.install(bundle_dir, source_dir)



# SPDX-FileCopyrightText: 2025 René de Hesselle <dehesselle@web.de>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import logging
import argparse
from pathlib import Path
from enum import Enum

from abcreate.bundle import Bundle
from abcreate.util.log import setup_logging, logstats

try:
    from abcreate._version import version
except ImportError:
    version = "0.0.0"

log = logging.getLogger("main")


class Command(Enum):
    CREATE = "create"


def main() -> None:
    parser = argparse.ArgumentParser(description="create an application bundle")
    parser.add_argument("--version", action="version", version=f"abcreate {version}")
    p_commands = parser.add_subparsers(help="available commands", dest="command")

    p_create = p_commands.add_parser(
        Command.CREATE.value, help="create application bundle"
    )
    p_create.add_argument("file", type=Path, help="XML configuration file")
    p_create.add_argument(
        "-i",
        "--install_prefix",
        type=Path,
        required=True,
        help="install prefix of the application",
    )
    p_create.add_argument(
        "-o",
        "--output_dir",
        type=Path,
        required=True,
        help="directory to create the .app bundle in",
    )

    args = parser.parse_args()

    if args.command == Command.CREATE.value:
        setup_logging("abcreate.log")
        log.info(f"abcreate {version}")

        try:
            xml_doc = args.file.read_text()
            bundle = Bundle.from_xml(xml_doc)
        except Exception as e:
            log.critical(e)
        else:
            bundle.create(args.output_dir, args.install_prefix)

    else:
        parser.print_usage()

    log.info(f"finished with {logstats.warnings} warnings and {logstats.errors} errors")

    if logstats.errors:
        exit(1)



# SPDX-FileCopyrightText: 2025 René de Hesselle <dehesselle@web.de>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import logging
from pathlib import Path

log = logging.getLogger("path")


def path_relative_to(path: Path, part: str, include_part: bool = False) -> Path:
    try:
        index = path.parts.index(part)
        offset = 0 if include_part else 1
        return Path(*path.parts[index + offset :])
    except ValueError:
        log.error(ValueError)
        return path



# SPDX-FileCopyrightText: 2025 René de Hesselle <dehesselle@web.de>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import logging
from pathlib import Path
from typing import List

from pydantic_xml import BaseXmlModel, element

from .icon import Icon
from .plist import Plist

log = logging.getLogger("icon")


class Icons(BaseXmlModel):
    icons: List[Icon] = element(tag="icon")

    @property
    def main_icon(self) -> Icon:
        try:
            return self.icons[0]
        except IndexError:
            log.critical("no icons specified")
            return None

    def install(self, bundle_dir: Path, install_prefix: Path):
        for icon in self.icons:
            icon.install(bundle_dir, install_prefix)

        Plist().CFBundleIconFile = Path(self.main_icon.source_path).name



# SPDX-FileCopyrightText: 2025 René de Hesselle <dehesselle@web.de>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import logging
from pathlib import Path
from typing import Optional
from shutil import copy

from pydantic_xml import BaseXmlModel, attr

from abcreate.util import LinkedObject
from .library import Library

log = logging.getLogger("executable")


class Executable(BaseXmlModel):
    name: Optional[Path] = attr(default=None)
    source_path: Path

    @property
    def target_name(self) -> str:
        return (self.name or self.source_path).name

    def install(self, bundle_dir: Path, install_prefix: Path):
        target_dir = bundle_dir / "Contents" / "MacOS"
        if not target_dir.exists():
            target_dir.mkdir(parents=True)

        if (source_path := install_prefix / "bin" / self.source_path).exists():
            target_path = target_dir / self.target_name
            if target_path.exists():
                log.error(f"will not overwrite {target_path}")
            else:
                log.debug(f"copy {source_path} to {target_path}")
                copy(source_path, target_path)

                # pull in dependencies
                linked_object = LinkedObject(source_path)
                if not linked_object.rpaths:
                    log.debug(f"assuming lib as default rpath for {source_path}")
                    linked_object.resolved_rpaths[install_prefix / "lib"] = None
                for path in linked_object.flattened_dependency_tree(
                    exclude_system=True
                ):
                    library = Library(source_path=path)
                    if library.is_framework:
                        # frameworks are taken care of separately
                        log.debug(
                            f"intentionally skipping framework library {library.source_path}"
                        )
                        pass
                    else:
                        library.install(bundle_dir, install_prefix)

                # adjust install names
                frameworks_dir = bundle_dir / "Contents" / "Frameworks"
                linked_object = LinkedObject(target_path)
                linked_object.change_dependent_install_names(
                    Path("@executable_path/../Frameworks"),
                    frameworks_dir,
                )
        else:
            log.error(f"cannot locate {self.source_path}")



# SPDX-FileCopyrightText: 2025 René de Hesselle <dehesselle@web.de>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import logging
from pathlib import Path
from shutil import rmtree
from typing import Optional

from pydantic import model_validator
from pydantic_xml import BaseXmlModel, element

from .executables import Executables
from .frameworks import Frameworks
from .gtk import Gir, Gtk3, Gtk4
from .icons import Icons
from .libraries import Libraries
from .locales import Locales
from .plist import Plist
from .resources import Resources

log = logging.getLogger("bundle")


class Bundle(BaseXmlModel, tag="bundle"):
    executables: Executables
    frameworks: Optional[Frameworks] = element(default=None)
    gir: Gir
    gtk3: Optional[Gtk3] = element(default=None)
    gtk4: Optional[Gtk4] = element(default=None)
    icons: Icons
    libraries: Optional[Libraries] = element(default=None)
    locales: Locales
    plist: Plist
    resources: Resources

    @model_validator(mode="after")
    def ensure_gtk3_gtk4_mutually_exclusive(self):
        if (self.gtk3 and self.gtk4) or (not self.gtk3 and not self.gtk4):
            log.critical("gtk3 and gtk4 are mutually exclusive")
        return self

    def create(self, output_dir: Path, install_prefix: Path):
        bundle_dir = output_dir / Path(
            self.executables.main_executable.target_name
        ).with_suffix(".app")

        if bundle_dir.exists():
            log.debug(f"removing {bundle_dir.as_posix()}")
            rmtree(bundle_dir)

        log.info(f"creating {bundle_dir.as_posix()}")
        bundle_dir.mkdir(parents=True)

        # It's important to install the plist first because others might
        # depend on it. (There is no dependency management.)
        log.info("--- {:>11} ---".format("plist"))
        self.plist.install(bundle_dir, install_prefix)
        if self.gtk3:
            log.info("--- {:>11} ---".format("gtk3"))
            self.gtk3.install(bundle_dir, install_prefix)
        if self.gtk4:
            log.info("--- {:>11} ---".format("gtk4"))
            self.gtk4.install(bundle_dir, install_prefix)
        log.info("--- {:>11} ---".format("gir"))
        self.gir.install(bundle_dir, install_prefix)
        if self.libraries:
            log.info("--- {:>11} ---".format("libraries"))
            self.libraries.install(bundle_dir, install_prefix)
        if self.frameworks:
            log.info("--- {:>11} ---".format("frameworks"))
            self.frameworks.install(bundle_dir, install_prefix)
        log.info("--- {:>11} ---".format("executables"))
        self.executables.install(bundle_dir, install_prefix)
        log.info("--- {:>11} ---".format("icons"))
        self.icons.install(bundle_dir, install_prefix)
        log.info("--- {:>11} ---".format("locales"))
        self.locales.install(bundle_dir, install_prefix)
        log.info("--- {:>11} ---".format("resources"))
        self.resources.install(bundle_dir, install_prefix)



# SPDX-FileCopyrightText: 2025 René de Hesselle <dehesselle@web.de>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import logging
from pathlib import Path
from shutil import copy

from pydantic_xml import BaseXmlModel

from abcreate.util import path_relative_to

log = logging.getLogger("locale")


class Locale(BaseXmlModel):
    name: str

    def install(self, bundle_dir: Path, install_prefix: Path):
        target_dir = bundle_dir / "Contents" / "Resources" / "share" / "locale"

        for source_path in Path(install_prefix / "share" / "locale").rglob(self.name):
            target_path = target_dir / path_relative_to(source_path, "locale")
            if target_path.exists():
                pass
                # log.debug(f"will not overwrite {target_path}")
            else:
                log.debug(f"copy {source_path} to {target_path}")
                target_path.parent.mkdir(parents=True, exist_ok=True)
                copy(source_path, target_path)



# SPDX-FileCopyrightText: 2025 René de Hesselle <dehesselle@web.de>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import logging
from pathlib import Path
from typing import List

from pydantic_xml import BaseXmlModel, element

from .resource import Resource

log = logging.getLogger("resource")


class Resources(BaseXmlModel):
    resources: List[Resource] = element(tag="resource")

    def install(self, bundle_dir: Path, install_prefix: Path):
        for resource in self.resources:
            resource.install(bundle_dir, install_prefix)
